/* -*- Mode: C; tab-width: 4; c-basic-offset: 4; indent-tabs-mode: nil -*- */
/*
 * mirror.c: Mirroring
 *
 * Copyright (c) 2000, 2001 Ximian, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */

#include <gnome.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <libsoup/soup-uri.h>

#include "mirror.h"
#include "rc-util.h"
#include "gui-transfer.h"
#include "gui-util.h"
#include "gui-prefs.h"
#include "xml-util.h"

typedef struct _RCMirror RCMirror;

struct _RCMirror {
    char *name;
    char *location;
    char *url;
};

static GSList *
parse_mirrors_xml(void)
{
    xmlDoc *doc;
    xmlNode *root, *node;
    GSList *mirrors = NULL;
    RCMirror *m;
    char *mirror_file;

    m = g_new0(RCMirror, 1);
    m->name = g_strdup(_("No mirror (always use default host)"));
    m->location = g_strdup(_("N/A"));
    m->url = NULL;
    mirrors = g_slist_append(mirrors, m);

    if (appinfo->not_root) {
        mirror_file = g_strdup_printf(
            "%s/.redcarpet/mirrors.xml", getenv("HOME"));
    }
    else
        mirror_file = g_strdup("/var/lib/redcarpet/mirrors.xml");

    if (!rc_file_exists(mirror_file)) {
        /* No mirrors.xml file on the local system yet. No big deal. */
        g_free(mirror_file);
        return mirrors;
    }

    doc = xmlParseFile(mirror_file);
    g_free(mirror_file);
    if (!doc) {
        GtkWidget *message;

        message = gnome_app_error(
            GNOME_APP(GET_WIDGET("main_window")), 
            _("Unable to parse mirror list."));
        gnome_dialog_run(GNOME_DIALOG(message));
        return NULL;
    }
#if LIBXML_VERSION < 20000
    root = doc->root;
#else
    root = xmlDocGetRootElement(doc);
#endif
    if (!root) {
        /* Uh. Bad. */
        rc_debug (RC_DEBUG_LEVEL_WARNING, "mirrors.xml document has no root\n");
        xmlFreeDoc(doc);
        return NULL;
    }

#if LIBXML_VERSION < 20000
    node = root->childs;
#else
    node = root->children;
#endif
    while (node) {
        char *tmp;
        RCMirror *mirror;

        tmp = xml_get_value(node, "internal");

        if (tmp && !getenv("RC_INTERNAL_MIRRORS")) {
            g_free(tmp);
            node = node->next;
            continue;
        }

        mirror = g_new0(RCMirror, 1);
        mirror->name = xml_get_value(node, "name");
        mirror->location = xml_get_value(node, "location");
        mirror->url = xml_get_value(node, "url");

        mirrors = g_slist_append(mirrors, mirror);

        node = node->next;
    }

    xmlFreeDoc(doc);

    return mirrors;
} /* parse_mirrors_xml */

static void
save_mirror_list(GByteArray *data)
{
    char *path;
    char *mirror_file;
    int fd;

    if (!data) {
        /* This must mean that an error occurred trying to download mirrors.xml.
           In this case, we'll just give up and leave whatever was there
           previously. We don't want to truncate a perfectly good mirrors.xml
           because the user kicked out the phone plug... */
        return;
    }

    if (appinfo->not_root) {
        char *home = getenv("HOME");

        path = g_strdup_printf("%s/.redcarpet", home);
    }
    else
        path = g_strdup("/var/lib/redcarpet");

    if (!g_file_exists(path)) {
        if (rc_mkdir(path, 0755))
            rc_debug(
                RC_DEBUG_LEVEL_WARNING, 
                "Unable to create library directory %s\n", path);
    }

    mirror_file = g_strdup_printf("%s/mirrors.xml", path);
    g_free(path);

    fd = open(mirror_file, O_CREAT | O_TRUNC | O_WRONLY, 0644);
    if (fd < 0) {
        char *m;
        GtkWidget *message;
        
        m = g_strdup_printf(
            _("Unable to save mirror list (%s)."), mirror_file);
        message = gnome_app_error(GNOME_APP(GET_WIDGET("main_window")), m);
        gnome_dialog_run(GNOME_DIALOG(message));
        g_free(m);
        g_free(mirror_file);
        return;
    }
    g_free(mirror_file);
    write(fd, data->data, data->len);
    close(fd);
} /* save_mirror_list */

static void
refresh_mirror_list(void)
{
    GSList *mirror_list = NULL;
    GSList *i;
    GtkCList *clist;

    mirror_list = parse_mirrors_xml();

    /* The "No mirror" option is always first. */
    if (appinfo->priority) {
        RCMirror *m = mirror_list->data;

        g_free(m->name);
        g_free(m->location);
        g_free(m);

        mirror_list = mirror_list->next;
    }

    clist = GTK_CLIST(GET_WIDGET("mirror_clist"));

    gtk_clist_clear(clist);
    i = mirror_list;
    while (i) {
        RCMirror *mirror = i->data;
        char *stuff[] = { mirror->name, mirror->location, NULL};
        int row;
        const char *selected_url;

        row = gtk_clist_append(clist, stuff);
        gtk_clist_set_row_data(clist, row, mirror->url);

        selected_url = gui_prefs_get_mirror_url();
        if (selected_url == mirror->url ||
            (selected_url && mirror->url &&
             g_strcasecmp(selected_url, mirror->url) == 0))
            gtk_clist_select_row(clist, row, 0);

        g_free(mirror);

        i = i->next;
    }
    gtk_clist_columns_autosize(clist);
    g_slist_free(mirror_list);
} /* refresh_mirror_list */

void
rc_gui_mirrors_fetch(void)
{
    GByteArray *data = NULL;
    GUITransfer *t;
    char *url;

    t = gui_transfer_new(GUI_TRANSFER_FLAGS_BLOCK, appinfo->normal_cache);
    if (appinfo->splash) {
        gui_transfer_set_progress_bar(
            t, GTK_PROGRESS(e_splash_get_progress_bar(appinfo->splash)));
    }
    else {
        gui_transfer_set_progress_bar(
            t, GTK_PROGRESS(GET_WIDGET("single_progress")));
        gui_transfer_set_progress_label(
            t, GTK_LABEL(GET_WIDGET("progress_label")));
    }

    url = g_strdup_printf("%s/mirrors.xml", rc_gui_get_host());
    data = gui_transfer_begin(t, url);
    g_free(url);

    if (gui_transfer_get_error(t)) {
        gboolean retry;

        retry = gui_transfer_error_box(
            t,
            _("<p>You can continue without any mirrors by "
              "pressing \"Cancel\""));

        if (retry) {
            if (data)
                g_byte_array_free(data, FALSE);
            return rc_gui_mirrors_fetch();
        }
        else {
            /* Oh well. */
            return;
        }
    }
    gtk_object_unref(GTK_OBJECT(t));

    save_mirror_list(data);
    g_byte_array_free(data, TRUE);
    refresh_mirror_list();
} /* rc_gui_mirrors_fetch */

void
rc_gui_mirrors_clear(void)
{
    gtk_clist_clear(GTK_CLIST(GET_WIDGET("mirror_clist")));
} /* rc_gui_mirrors_clear */

char *
rc_gui_get_mirrored_url(const char *file, RCChannel *rc)
{
    const char *mirror_url;
    gboolean mirrored;

    g_return_val_if_fail(file, NULL);

    if (rc_url_is_absolute(file))
        return g_strdup(file);

    mirror_url = gui_prefs_get_mirror_url();
    if (mirror_url && *mirror_url) {
        if (rc && !rc->mirrored)
            mirrored = FALSE;
        else
            mirrored = TRUE;
    }
    else
        mirrored = FALSE;

    return g_strdup_printf(
        "%s/%s", mirrored ? mirror_url : rc_gui_get_host(),
        file[0] == '/' ? file + 1 : file);
} /* rc_gui_get_mirrored_url */
